/*
 * Copyright 1991 Addison-Wesley Publishing Company
 *                (see the file "COPYRIGHTS" for details)
 */

#include "motifhelp.h"
#include <stdio.h>

/************************** Utility *********************************/

XmString MakeXmString (string)
    char    *string;
{
    return XmStringLtoRCreate (string, XmSTRING_DEFAULT_CHARSET);
}

/************************* Info Storage *****************************/

/* list of class names as XmStrings for list display and selection */

#define MAX_CLASS_NAMES 50

static XmString class_name_list[MAX_CLASS_NAMES];
static int      no_class_names = 0;

/* root of class tree */

ClassDoc *root_class = NULL;

/************************* Sorting ***********************************/

#define SORTSIZE 400

static XmString sortbuf[SORTSIZE];
int sort_entries;
int toggle = 0;   /* to deliver different lists to set values */

void InsertItem (item)
    XmString    item;
{
    int         pos, i;
    String      new, old;

    XmStringGetLtoR (item, "", &new);
    for (pos = toggle; pos < sort_entries+toggle; pos++){
        XmStringGetLtoR (sortbuf[pos], "", &old);
        if (strcmp (old, new) >= 0)
            break;
    }
    for (i = sort_entries+toggle; i > pos; i--)
        sortbuf [i] = sortbuf[i-1];
    sort_entries++;
    sortbuf[pos] = item;
}
    
XmString *SortStringList (list, no)
    XmString    list[];
    int         no;
{
    int         i;

    if (toggle == 0) toggle = 1; else toggle = 0;
    sort_entries = 0;
    for (i=0; i<no; i++)
        InsertItem (list[i]);
    return &sortbuf[toggle];
}


/*************************** Help Dir *********************************/

#define NAME_BUFLENGTH 200
#define MHELP_PREFIX "motifhelp."

String dir;

void SetHelpDir (new_dir)
    String    new_dir;
{
    dir = new_dir;
}

FILE *help_open (filename)
    String    filename;
{
    char      namebuffer[NAME_BUFLENGTH];

    strcpy (namebuffer, dir);
    strcat (namebuffer, MHELP_PREFIX);
    strcat (namebuffer, filename);
    return fopen (namebuffer, "r");
}

/************************* Initialization *****************************/

/* allocate record for new class during init */

ClassDoc *AllocClassRec (name, parent)
    String    name;
    ClassDoc *parent;
{
    ClassDoc *me = XtNew (ClassDoc);
    int       i;

    me->name = XtMalloc (strlen (name) + 1);
    strcpy (me->name, name);
    me->xm_name = MakeXmString (name);
    me->super = parent;
    if (parent) {
        parent->subs[parent->no_subs] = me;
        parent->no_subs++;
        me->level = parent->level + 1; }
    else
        me->level = 1;
    me->no_subs = 0;
    me->no_res = 0;
    me->sections = 0;
    me->file_read = False;
    return me;
}

/* initialize list of classes and allocate records */

#define BIG_BUFLENGTH 1000

void ReadClassNameList ()
{
    FILE       *in;
    char        inbuffer[BIG_BUFLENGTH];
    int         newlinepos;
    ClassDoc   *doc = NULL;
    int         indent, prev_indent= -1;
    XmString   *sortlist;
    int         i;

    in = help_open ("classes");
    if (in == NULL)
        XtErrorMsg (
            NErrFileOpen, TErrClassfile, /* resource name         */
            ERROR_CLASS,                 /* resource class        */
            "cannot open motifhelp.classes",  /* default message  */
            NULL, NULL);                 /* parameters            */
    no_class_names = 0;
    while (fgets (inbuffer, BIG_BUFLENGTH, in)) {
        /* strip trailing newline */
        newlinepos = strlen (inbuffer)-1;
        if (inbuffer[newlinepos] == '\n') inbuffer[newlinepos] = '\0';
        /* count indentation as level in hierarchy */
        indent=0;
        while (inbuffer[indent] == ' ') indent++;
        /* determine parent by comparing with previous indent */
        while (prev_indent-- >= indent) doc = doc->super;
        prev_indent = indent;
        /* allocate */
        doc = AllocClassRec (&inbuffer[indent], doc);
        if (doc->super == NULL) root_class = doc;
        class_name_list[no_class_names++] = doc->xm_name;
    }
    fclose (in);
    sortlist = SortStringList (class_name_list, no_class_names);
    for (i = 0; i < no_class_names; i++)
        class_name_list [i] = sortlist [i];
}

/* external interface procedures */

XmString *GetClassNames (no_of_names)
    int        *no_of_names;
{
    if (no_class_names == 0) ReadClassNameList ();
    *no_of_names = no_class_names;
    return class_name_list;
}

ClassDoc * RootClass ()
{
    if (no_class_names == 0) ReadClassNameList ();
    return root_class;
}

/*********************** Searching ************************************/

/* recursive search procedure in class tree */

ClassDoc *SearchClassTree (node, name)
    ClassDoc   *node;
    XmString    name;
{
    ClassDoc    *res;
    int          i;

    if (XmStringCompare (node->xm_name, name)) return node;
    for (i=0; i<node->no_subs; i++)
        if ((res = SearchClassTree (node->subs[i], name)) != NULL)
            return res;
    return NULL;
}    

/* search class tree for named (XmString!) class */
   
ClassDoc *SearchClass (name)
    XmString    name;
{
    return SearchClassTree (root_class, name);
}

/******************************* File Scanning ***************************/

#define BIGBIG_BUFSIZE 40000
#define MAXLINE 200
#define BIGLIMIT BIGBIG_BUFSIZE-MAXLINE

#define MAX_FILENAME 50
#define SMALL_BUFLENGTH 130

#define BLANK ' '
#define TAB '\t'
#define NULLC '\0'
#define NEWLINE '\n'

char  bigbuffer[BIGBIG_BUFSIZE];
char  inbuffer[SMALL_BUFLENGTH];

/* test for containment */

Boolean IsIn (in, str)
    char   *in;
    char   *str;
{
    char   *current = in;

    while (*current != NULLC) {
        if (*current == *str)
            if (strncmp (current, str, strlen (str)) == 0)
                return True;
        current++; 
    }
    return False;
}

/* find range of following fields from header */

int FindRange (buf)
    char   *buf;
{
    if (IsIn (buf, "Constraint")) return 1;
    if (IsIn (buf, "Scrolled")) return 2;
    if (IsIn (buf, "Special")) return 3;
    return 0;
}

/* extract next table field between tabs */

char *NextInfield (start_pos)
    int    *start_pos;
{
    char   *start = &inbuffer[*start_pos];
    char   *current = start;
    char    last;
    int     len=0;

    while (*current != NEWLINE && *current != TAB && *current != NULLC)
        {current++; len++;}
    last = *current;
    *current = NULLC;
    if (last != NULLC) (*start_pos) += len + 1;
    return start;
}

/* Read Single Line Entry in Class Table */

void ReadClassTableEntry (doc, range)
    ClassDoc  *doc;
    int        range;
{
    int        blankpos, entry, newl, val;
    char      *new;

    blankpos = 6;
    entry = doc->no_res;
    new = NextInfield (&blankpos);
    newl = strlen (new) + 1;
    doc->res_name[entry] = XtMalloc (newl);
    strcpy (doc->res_name[entry], new);
    doc->res_string[entry] =
        MakeXmString (doc->res_name[entry]);
    doc->res_class[entry] =
        MakeXmString (NextInfield (&blankpos));
    doc->res_type[entry] =
        MakeXmString (NextInfield (&blankpos));
    doc->res_default[entry] =
        MakeXmString (NextInfield (&blankpos));
    new = NextInfield (&blankpos);
    val = 0;
    if (*new == 'C') {val +=1; new++;}
    if (*new == 'S') {val +=2; new++;}
    if (*new == 'G') {val +=4; new++;}
    doc->res_access[entry] = val;
    doc->res_range[entry] = range;
    doc->no_res++;
}

/* read long text section */

void ReadLongText (file)
    FILE        *file;
{
    int          count=0;
    while (fgets (&bigbuffer[count], MAXLINE, file)) {
        if (bigbuffer[count] == '#') {
            strcpy (inbuffer, &bigbuffer[count]);
            bigbuffer[count] = NULLC;
            return; 
	}
        if (count < BIGLIMIT) count += strlen (&bigbuffer[count]);
    }
    inbuffer[0] = NULLC;
}

/* read one section of file */

Boolean ReadResourceText (doc, in)
    ClassDoc  *doc;
    FILE      *in;
{
    char      *name = &inbuffer[6];
    int        findex, no_params = 1;
    char      *nameend = &inbuffer[strlen(inbuffer)-1];

    while (*nameend == NEWLINE || *nameend == TAB || *nameend == BLANK)
        nameend--;
    *(++nameend) = NULLC; 
    for (findex=0; findex < doc->no_res; findex++)
        if (strcmp (name, doc->res_name[findex]) == 0) break;
    if (findex == doc->no_res) {
        findex = 1;
        XtWarningMsg (
            NWarnNFnd, TWarnField,    /* resource name         */
            WARN_CLASS,               /* resource class        */
            "Field %s not found in table", /* default message  */
            &name, &no_params);       /* parameters            */
        return False;
    }
    ReadLongText (in);
    doc->res_text[findex] = XtMalloc (strlen (bigbuffer) + 1);
    strcpy (doc->res_text[findex], bigbuffer);
    return True;
}
    

/* compare which sections exist */

#define SECTION0 "DESCRIPTION"
#define SECTION1 "Behavior"
#define SECTION2 "Callback"
#define SECTION3 "Default Translation"
#define SECTION4 "Traversal"
#define SECTION5 "Color"

#define NUM_SECTIONS 6

static String section_headers[] =
    {SECTION0, SECTION1, SECTION2, SECTION3, SECTION4, SECTION5};

/* test if one of the sections is present */

void CompareSection (doc)
    ClassDoc *doc;
{
    int       i, mask = 1;

    for (i=0; i<NUM_SECTIONS; i++){
        if (strncmp (&inbuffer[3], section_headers[i],
                     strlen (section_headers[i])) == 0)
            doc->sections += mask;
        mask += mask;
    }
}

/* read info text of class on demand from separate file */

void FillHelpInfo (doc)
    ClassDoc  *doc;
{
    FILE      *in;
    Boolean    res_area, ahead;
    int        newlinepos, range = 0, no_params = 1;
    char      *getres;

    /* test if already read */
    if (doc->file_read) return;
    doc->file_read = True;

    /* filename = motifhelp.<classname> */
    in = help_open (doc->name);
    if (in == NULL) {
        range = 1;
        XtWarningMsg (
            NWarnNFnd, TWarnFile,      /* resource name         */
            WARN_CLASS,                /* resource class        */
            "File %s not found",       /* default message       */
            &(doc->name), &no_params); /* parameters            */
        return; /* ignore for now */
    }

    res_area = False;
    getres = fgets (inbuffer, SMALL_BUFLENGTH, in);
    while (getres) {
        ahead = False;
        if (res_area) {
            if (strncmp (inbuffer, "#1 ", 3) == 0)
                res_area = False;
            if (strncmp (inbuffer, "#2 ", 3) == 0) 
                ahead = ReadResourceText (doc, in); 
            if (strncmp (inbuffer, "#3 ", 3) == 0)
                 ReadClassTableEntry (doc, range);
            if (strncmp (inbuffer, "#4 ", 3) == 0)
                range = FindRange (inbuffer);}
        else {
            if (strcmp (inbuffer, "#1 New Resources\n") == 0) res_area = True;
            if (strncmp (inbuffer, "#1 ", 3) == 0) CompareSection (doc);
	}
        if (!ahead) getres = fgets (inbuffer, SMALL_BUFLENGTH, in);
    }
}

/* read long section into bigbuffer */

String ReadLongSection (ending, section)
    char      *ending;
    char      *section;
{
    FILE      *in;
    char       comp[MAX_FILENAME];
    int        complen, no_params = 1;

    /* filename = motifhelp.<classname> */
    in = help_open (ending);
    bigbuffer[0] = NULLC;
    if (in == NULL) {
        complen = 1;
        XtWarningMsg (
            NWarnNFnd, TWarnFile,     /* resource name         */
            WARN_CLASS,               /* resource class        */
            "File %s not found",      /* default message       */
            &ending, &no_params);     /* parameters            */
        return; /* ignore for now */
    }

    strcpy (comp, "#1 ");
    strcat (comp, section);
    complen = strlen (comp);
    while (fgets (inbuffer, SMALL_BUFLENGTH, in)) {
        if (strncmp (inbuffer, comp, complen) == 0)
            {ReadLongText (in); return bigbuffer;}
    }
    return bigbuffer;
}

/* read special labeled section from manpage into bigbuffer */

String ReadSection (doc, section)
    ClassDoc  *doc;
    int        section;
{
    if (section >= 1 && section <= NUM_SECTIONS)
      return ReadLongSection (doc->name, section_headers [section-1]);
    else
        XtErrorMsg (
            NErrInternal, TErrSection, /* resource name         */
            ERROR_CLASS,               /* resource class        */
            "wrong section number",    /* default message       */
            NULL, NULL);               /* parameters            */
}


/****************************** Field Searching **************************/

#define MAX_MATCH 300

struct {
    ClassDoc *doc;
    int       entry;
    int       multiple;
} m_res[MAX_MATCH];         /* table of matching resource fields */

static XmString show[MAX_MATCH];

static int no_res_match = 0;

/* find index of selected resource in match table */

void LocateResource (xm, doc, entry)
    XmString      xm;
    ClassDoc    **doc;
    int          *entry;
{
    int           i;

    *doc = NULL;
    for (i=0; i<no_res_match; i++)
        if (XmStringCompare (xm, show[i])) {
            *doc = m_res [i]. doc;
            *entry = m_res [i]. entry;
            return; 
        }
}

/* recursively search classes and assemble match table */

void SearchResource (name, doc)
    String        name;
    ClassDoc     *doc;
{
    int           i, j;

    FillHelpInfo (doc);
    for (i=0; i<doc->no_res; i++)
        if (IsIn (doc->res_name[i], name)) {
            m_res[no_res_match].doc = doc;
            m_res[no_res_match].entry = i;
            m_res[no_res_match].multiple = -1;
            show[no_res_match] = NULL;
            for (j=no_res_match-1; j>=0; j--)
                if (strcmp (m_res[j].doc->res_name[m_res[j].entry],
                            doc->res_name[i]) == 0)
		    {m_res[no_res_match].multiple=j; break;}
            no_res_match++;
	}
    for (i=0; i<doc->no_subs; i++)
       SearchResource (name, doc->subs[i]);
}           

/* set representation of matching resource (append classname) */

void SetShow (index, multi)
    int           index;
    Boolean       multi;
{
    char          buffer[SMALL_BUFLENGTH];

    strcpy (buffer, m_res[index].doc->res_name[m_res[index].entry]);
    if (multi) {
        strcat (buffer, " <");
        strcat (buffer, m_res[index].doc->name);
        strcat (buffer, ">");
    }
    show[index] = MakeXmString (buffer);
}

/* create match table for incomplete specification */

void FindMatchingFields (name, list, no_fields)
    String        name;
    XmString    **list;
    int          *no_fields;
{
    int           i, j;

    for (i=0; i<no_res_match; i++) XmStringFree (show[i]);
    no_res_match = 0;
    SearchResource (name, root_class);
    for (i=no_res_match-1; i>=0; i--)
        if (show[i] == NULL) 
            if (m_res[i].multiple >= 0) {
                j = i;
                while (j>=0) {SetShow (j, True); j=m_res[j].multiple;}
	    }
            else
                SetShow (i, False);
    *no_fields = no_res_match;
    *list = SortStringList (show, no_res_match);
}



